#!/usr/bin/python3
"""Example driver for Samsara.

To create a samsara driver for your own project, copy this file and
add in your function and options.

Checklist
---------

A small checklist of things to change before use:
  - Either overwrite the `objective` function or import your function.
  - If you imported your own function change the `Objective` variable
in the function `driver` to your function.
  - Optional: Fill `plot_result` with your plotting routine.
  - Change the starting value of the starting point `xold_vec`
  - Customise Samsara via the given options to your needs
  - Check tolerances for the main loop
"""
import matplotlib.pyplot as plt
from numpy import abs, amax, array, load, zeros
from numpy.linalg import norm

from samsara import Samsara, BFGS1_product, Broyden1_product,\
    Broyden2_product, Dogleg_QN, Explicit_TR, cdSY_mat, fdSY_mat


def objective(x):
    """Example objective (Engvall function).

    Replace this code with your actual function.

    Parameters
    ----------
    x : array_like
        The point to evaluate the function at (here 2 dimensional).

    Returns
    -------
    f : float
        The function value at `x`.
    Df : array_like
        The gradient vector of `f` at `x`.

    Notes
    -----
    In your implementation you may use any number of parameters and
    return values. But samsara will always need the function value
    and the gradient vector.
    """
    f = x[0]**4 + x[1]**4 + 2*x[0]**2*x[1]**2 - 4*x[0] + 3
    Df = zeros(2)
    Df[0] = 4*x[0]**3 + 4*x[0]*x[1]**2 - 4
    Df[1] = 4*x[1]**3 + 4*x[0]**2*x[1]

    return f, Df


def plot_result():
    """Example of a custom plot function for the result.

    Replace this code with your actual plotting routine.
    """
    pass


def driver():
    # You can either implement your function in the objective function
    # provided above, or import your function from a different module
    # and set it here.
    Objective = objective

    # --- TOLERANCES ---
    # The lower bound for the gradient vector norm.
    ngrad_TOL = 6e-6   # eps**(1/3)
    # The lower bound for the stepsize.
    step_TOL = 4e-11   # eps**(2/3)
    # The maximum number of iterations.
    Maxit = 500

    # ========== SAMSARA ==========
    # --- OPTIONS ---
    options = {}
    # Extra output from samsara.
    options['verbose'] = True
    # A regularisation parameter.
    options['alpha_0'] = 5e-12
    # ...
    options['gamma'] = .5
    # ...
    options['c'] = .01
    # ...
    options['beta'] = .9999999
    # ...
    options['eta1'] = .995
    # ...
    options['eta2'] = .8
    # ...
    options['eta3'] = .25
    # The maximum steps to look back. Must be an integer.
    options['maxmem'] = 8
    # The initial radius of the trust region.
    options['tr'] = 1e+15
    # Gradient norm tolerance
    # This is used as a stopping criteria for SAMSARA. It must be a
    # real number greater than zero.
    options['ngrad_TOL'] = ngrad_TOL
    # Step size tolerance
    # This is used as a stopping criteria for SAMSARA. It must be a
    # real number greater than zero.
    options['step_TOL'] = step_TOL
    # Maximum number of iterations
    # This is used as a stopping criteria for SAMSARA. It must be
    # an integer greater than zero.
    options['Maxit'] = Maxit
    # Hessian method
    # Quasi-Newton methods for estimation of the function's Hessian.
    # Broyden implementations use Dogleg_QN lin search methods
    # while BFGS implementations use Explicit_TR.
    # Methods available:
    #     - BFGS1_product
    #     - Broyden1_product
    #     - Broyden2_product
    options['QN_method'] = BFGS1_product
    # Step finder
    # Optional: Default is Dogleg_QN
    # Methods available:
    #     - Dogleg_QN
    #     - Explicit_TR
    options['Step_finder'] = Explicit_TR
    # History
    # Ordering methods for S, Y matrices in limited memory application.
    # Finite difference ordering, fdSY_mat, is recommended for Broyden
    # implementations. For BFGS implementations, use conjugate
    # direction ordering, cdSY_mat.
    options['History'] = cdSY_mat
    # Trust Region Adjust
    # Select the method for adjustment of the trust region in optimazation.
    options['update_conditions'] = 'Trust Region'
    # Initial Step Scaler
    # The norm of the initial step taken in the Cauchy direction.
    # This multiplied against the normalized gradient to yield the
    # initial direction vector in order to generate the first step
    # taken by SAMSARA.
    # Assign a value of 0.0D0 to use the default value which is the
    # minimum of 1D-1 or the norm(1d-1*gradient).
    options['initial_step'] = 1e+5*options['step_TOL']

    samsara = Samsara(**options)

    # ========== STARTING VALUES ==========
    # The starting vector. Set it to an appropriate value.
    xold_vec = array([.5, 2.])
    # The starting function value and gradient vector.
    fold, gradfold_vec = Objective(xold_vec)

    # The next point.
    xnew_vec = None
    # The next function value.
    fnew = None
    # The next gradient vector.
    gradfnew_vec = None
    # Keep at this value, so that the main loop executes at least once.
    stepsize = 999.
    # Keep at this value, so that the main loop executes at least once.
    ngradfnew = 999.

    # ========== MAIN LOOP ==========
    # Number of current iterations
    it = 0

    while it < Maxit and ngradfnew > ngrad_TOL and stepsize > step_TOL:
        xnew_vec, xold_vec, fold, gradfold_vec, stepsize =\
            samsara.run(xold_vec, xnew_vec, fold, fnew, gradfold_vec,
                        gradfnew_vec)
        it += 1
        fnew, gradfnew_vec = Objective(xnew_vec)
        ngradfnew = norm(gradfnew_vec)

    if stepsize <= step_TOL:
        print('Algorithm stagnated:  stepsize tolerance violated.')
    if it >= Maxit:
        print('Algorithm exceeded:  maximum step count violated.')

    print('iterations:', it, '; optimal value:', fnew)

    # TODO(benedikt) Rework from here
    # You can save some of the values of samsara.
    # samsara.save()

    # You can now plot your results using your custom function or
    # Samsara's built-in one.
    # plot_result()
    samsara.plot(fopt=fnew)

    # Return the optimal values.
    return xnew_vec, fnew


# Entry point into the driver.
if __name__ == '__main__':
    xnew, fnew = driver()
