"""Tests all functions from hessian.py.
"""
import sys
import os
import matlab.engine
from numpy import abs, array, amax, random

# Add samsara to the path.
PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(
    os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

from samsara import *

# The upper boundary for the relative error between the MATLAB and
# Python implementations.
EPS = 1e-8
# The number of repetitions for each test.
ITERATIONS = 10

eng = matlab.engine.start_matlab()
eng.addpath('../../matlab')


def gen_params():
    """
    Generates random parameters for the tests.
    """
    dim = random.randint(2, 30)
    mem = random.randint(1, dim+1)
    u_vec = eng.random('unif', -100., 100., dim, 1)
    S_mat = eng.random('unif', 0., 100., dim, dim)
    Y_mat = eng.random('unif', 0., 100., dim, dim)
    scale = eng.random('unif', .1, 10.)
    alpha = eng.random('unif', .1, 10.)

    return dim, mem, u_vec, S_mat, Y_mat, scale, alpha


def test_BFGS1_product():
    for _ in range(ITERATIONS):
        dim, mem, u_vec, S_mat, Y_mat, scale, alpha = gen_params()
        res_mat = eng.BFGS1_product(0, 0, u_vec, S_mat, Y_mat, scale, 0)
        res_py = BFGS1_product(array(u_vec).flatten(), array(S_mat),
                               array(Y_mat), scale)
        diff = array(res_mat).flatten() - res_py.flatten()
        assert amax(abs(diff)) < EPS*amax(abs(array(res_mat).flatten()))


def test_invBFGS1_product():
    for _ in range(ITERATIONS):
        dim, mem, u_vec, S_mat, Y_mat, scale, alpha = gen_params()
        res_mat = eng.invBFGS1_product(0, 0, u_vec, S_mat, Y_mat, scale, alpha)
        res_py = invBFGS1_product(array(u_vec).flatten(),
                                  array(S_mat), array(Y_mat), scale)
        diff = array(res_mat).flatten() - res_py.flatten()
        assert amax(abs(diff)) < EPS*amax(abs(array(res_mat).flatten()))


def test_Broyden1_product():
    for _ in range(ITERATIONS):
        dim, mem, u_vec, S_mat, Y_mat, scale, alpha = gen_params()
        res_mat = eng.Broyden1_product(0, mem, u_vec, S_mat, Y_mat,
                                       scale, 0)
        res_py = Broyden1_product(mem, array(u_vec).flatten(),
                                  array(S_mat), array(Y_mat), scale)
        # Without flatten() we would have res_mat.shape = (x,1) and
        # res_py.shape = (x,) and the result of the subtraction would
        # be a matrix.
        diff = array(res_mat).flatten() - res_py.flatten()
        assert amax(abs(diff)) < EPS*amax(abs(array(res_mat).flatten()))


def test_invBroyden1_product():
    for _ in range(ITERATIONS):
        dim, mem, u_vec, S_mat, Y_mat, scale, alpha = gen_params()
        res_mat = eng.invBroyden1_product(0, dim, u_vec, S_mat, Y_mat,
                                          scale, alpha)
        res_py = invBroyden1_product(array(u_vec).flatten(),
                                     array(S_mat), array(Y_mat),
                                     scale, alpha)
        diff = array(res_mat).flatten() - res_py.flatten()
        assert amax(abs(diff)) < EPS*amax(abs(array(res_mat).flatten()))


def test_Broyden2_product():
    for _ in range(ITERATIONS):
        dim, mem, u_vec, S_mat, Y_mat, scale, alpha = gen_params()
        res_mat = eng.Broyden2_product(0, mem, u_vec, S_mat, Y_mat,
                                       scale, 0)
        res_py = Broyden2_product(mem, array(u_vec).flatten(),
                                  array(S_mat), array(Y_mat), scale)
        # Regarding flatten(): See comment in test_Broyden1_product.
        diff = array(res_mat).flatten() - res_py.flatten()
        assert amax(abs(diff)) < EPS*amax(abs(array(res_mat).flatten()))


def test_invBroyden2_product():
    for _ in range(ITERATIONS):
        dim, mem, u_vec, S_mat, Y_mat, scale, alpha = gen_params()
        res_mat = eng.invBroyden2_product(0, dim, u_vec, S_mat, Y_mat,
                                          scale, alpha)
        res_py = invBroyden2_product(array(u_vec).flatten(),
                                     array(S_mat), array(Y_mat),
                                     scale, alpha)
        diff = array(res_mat).flatten() - res_py.flatten()
        assert amax(abs(diff)) < EPS*amax(abs(array(res_mat).flatten()))


def test_delta_MSB2():
    for _ in range(ITERATIONS):
        dim, mem, _, S_mat, Y_mat, _, alpha = gen_params()
        res_mat = eng.delta_MSB2(dim, dim, S_mat, Y_mat, alpha)
        res_py = delta_MSB2(array(S_mat), array(Y_mat), alpha)
        diff = res_mat - res_py
        assert abs(diff) < EPS*abs(res_mat)


def test_delta_invBFGS():
    for _ in range(ITERATIONS):
        dim, mem, _, S_mat, Y_mat, _, alpha = gen_params()
        res_mat = eng.delta_invBFGS(dim, mem, S_mat, Y_mat, 0)
        res_py = delta_invBFGS(mem, array(S_mat), array(Y_mat))
        diff = res_mat - res_py
        assert abs(diff) < EPS*abs(res_mat)
